package wellington

import (
	"errors"
	"fmt"
	"io/ioutil"
	"path/filepath"
	"regexp"
	"strings"
)

// ImportPath accepts a directory and file path to find partials for importing.
// Returning a new pwd and string of the file contents.
// File can contain a directory and should be evaluated if
// successfully found.
// Dir is used to provide relative context to the importee.  If no file is found
// pwd is echoed back.
//
// Paths are looked up in the following order:
// {includepath}/_file.scss
// {includepath}/file.scss
// {Dir{dir+file}}/_{Base{file}}
// {Dir{dir+file}}/{Base{file}}
func (p *Parser) ImportPath(dir, file string) (string, string, error) {
	// fmt.Println("Importing: " + file)
	baseerr := ""
	//Load and retrieve all tokens from imported file
	path, _ := filepath.Abs(fmt.Sprintf("%s/%s.scss", dir, file))
	pwd := filepath.Dir(path)
	// Sass put _ in front of imported files
	fpath := filepath.Join(pwd, "/_"+filepath.Base(path))
	contents, err := ioutil.ReadFile(fpath)
	if err == nil {
		p.PartialMap.AddRelation(p.MainFile, fpath)
		return pwd, string(contents), nil
	}
	baseerr += fpath + "\n"
	if strings.HasSuffix(err.Error(), "no such file or directory") {
		// Look through the import path for the file
		for _, lib := range p.Includes {
			path, _ := filepath.Abs(lib + "/" + file)
			pwd := filepath.Dir(path)
			fpath = filepath.Join(pwd, "/_"+filepath.Base(path)+".scss")
			contents, err := ioutil.ReadFile(fpath)
			baseerr += fpath + "\n"
			if err == nil {
				p.PartialMap.AddRelation(p.MainFile, fpath)
				return pwd, string(contents), nil
			}
			// Attempt invalid name lookup (no _)
			fpath = filepath.Join(pwd, "/"+filepath.Base(path)+".scss")
			contents, err = ioutil.ReadFile(fpath)
			baseerr += fpath + "\n"
			if err == nil {
				p.PartialMap.AddRelation(p.MainFile, fpath)
				return pwd, string(contents), nil
			}
		}
	}
	// Ignore failures on compass
	re := regexp.MustCompile("compass\\/?")
	if re.Match([]byte(file)) {
		return pwd, string(contents), nil //errors.New("compass")
	}
	if file == "images" {
		return pwd, string(contents), nil
	}
	return pwd, string(contents), errors.New("Could not import: " +
		file + "\nTried:\n" + baseerr)
}
